/*
 * Minecraft Forge
 * Copyright (c) 2016.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation version 2.1
 * of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

package net.minecraftforge.fluids;

import net.minecraft.block.Block;
import net.minecraft.block.BlockDispenser;
import net.minecraft.block.material.Material;
import net.minecraft.creativetab.CreativeTabs;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Items;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.stats.StatList;
import net.minecraft.util.*;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.RayTraceResult;
import net.minecraft.util.text.translation.I18n;
import net.minecraft.world.World;
import net.minecraftforge.common.capabilities.ICapabilityProvider;
import net.minecraftforge.event.entity.player.FillBucketEvent;
import net.minecraftforge.fluids.capability.wrappers.FluidBucketWrapper;
import net.minecraftforge.fml.common.eventhandler.Event;
import net.minecraftforge.fml.common.eventhandler.EventPriority;
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;
import net.minecraftforge.items.ItemHandlerHelper;

import java.util.List;

/**
 * A universal bucket that can hold any liquid
 */
public class UniversalBucket extends Item implements IFluidContainerItem
{

    private final int capacity; // how much the bucket holds
    private final ItemStack empty; // empty item to return and recognize when filling
    private final boolean nbtSensitive;

    public UniversalBucket()
    {
        this(Fluid.BUCKET_VOLUME, new ItemStack(Items.field_151133_ar), false);
    }

    /**
     * @param capacity        Capacity of the container
     * @param empty           Item used for filling with the bucket event and returned when emptied
     * @param nbtSensitive    Whether the empty item is NBT sensitive (usually true if empty and full are the same items)
     */
    public UniversalBucket(int capacity, ItemStack empty, boolean nbtSensitive)
    {
        this.capacity = capacity;
        this.empty = empty;
        this.nbtSensitive = nbtSensitive;

        this.func_77625_d(1);

        this.func_77637_a(CreativeTabs.field_78026_f);

        BlockDispenser.field_149943_a.func_82595_a(this, DispenseFluidContainer.getInstance());
    }

    @SideOnly(Side.CLIENT)
    @Override
    public void func_150895_a(Item itemIn, CreativeTabs tab, List<ItemStack> subItems)
    {
        for (Fluid fluid : FluidRegistry.getRegisteredFluids().values())
        {
            if (fluid != FluidRegistry.WATER && fluid != FluidRegistry.LAVA && !fluid.getName().equals("milk"))
            {
                // add all fluids that the bucket can be filled  with
                FluidStack fs = new FluidStack(fluid, getCapacity());
                ItemStack stack = new ItemStack(this);
                if (fill(stack, fs, true) == fs.amount)
                {
                    subItems.add(stack);
                }
            }
        }
    }

    @Override
    public String func_77653_i(ItemStack stack)
    {
        FluidStack fluidStack = getFluid(stack);
        if (fluidStack == null)
        {
            if(getEmpty() != null)
            {
                return getEmpty().func_82833_r();
            }
            return super.func_77653_i(stack);
        }

        String unloc = this.func_77657_g(stack);

        if (I18n.func_94522_b(unloc + "." + fluidStack.getFluid().getName()))
        {
            return I18n.func_74838_a(unloc + "." + fluidStack.getFluid().getName());
        }

        return I18n.func_74837_a(unloc + ".name", fluidStack.getLocalizedName());
    }

    @Override
    public ActionResult<ItemStack> func_77659_a(ItemStack itemstack, World world, EntityPlayer player, EnumHand hand)
    {
        FluidStack fluidStack = getFluid(itemstack);
        // empty bucket shouldn't exist, do nothing since it should be handled by the bucket event
        if (fluidStack == null)
        {
            return ActionResult.newResult(EnumActionResult.PASS, itemstack);
        }

        // clicked on a block?
        RayTraceResult mop = this.func_77621_a(world, player, false);

        if(mop == null || mop.field_72313_a != RayTraceResult.Type.BLOCK)
        {
            return ActionResult.newResult(EnumActionResult.PASS, itemstack);
        }

        BlockPos clickPos = mop.func_178782_a();
        // can we place liquid there?
        if (world.func_175660_a(player, clickPos))
        {
            // the block adjacent to the side we clicked on
            BlockPos targetPos = clickPos.func_177972_a(mop.field_178784_b);

            // can the player place there?
            if (player.func_175151_a(targetPos, mop.field_178784_b, itemstack))
            {
                // try placing liquid
                if (FluidUtil.tryPlaceFluid(player, player.func_130014_f_(), fluidStack, targetPos)
                        && !player.field_71075_bZ.field_75098_d)
                {
                    // success!
                    player.func_71029_a(StatList.func_188057_b(this));

                    itemstack.field_77994_a--;
                    ItemStack emptyStack = getEmpty() != null ? getEmpty().func_77946_l() : new ItemStack(this);

                    // check whether we replace the item or add the empty one to the inventory
                    if (itemstack.field_77994_a <= 0)
                    {
                        return ActionResult.newResult(EnumActionResult.SUCCESS, emptyStack);
                    }
                    else
                    {
                        // add empty bucket to player inventory
                        ItemHandlerHelper.giveItemToPlayer(player, emptyStack);
                        return ActionResult.newResult(EnumActionResult.SUCCESS, itemstack);
                    }
                }
            }
        }

        // couldn't place liquid there2
        return ActionResult.newResult(EnumActionResult.FAIL, itemstack);
    }

    // compatibility
    @Deprecated
    public boolean tryPlaceFluid(Block block, World worldIn, BlockPos pos)
    {
        if (block instanceof IFluidBlock)
        {
            IFluidBlock fluidBlock = (IFluidBlock) block;
            return FluidUtil.tryPlaceFluid(null, worldIn, new FluidStack(fluidBlock.getFluid(), Fluid.BUCKET_VOLUME), pos);
        }
        else if (block.func_176223_P().func_185904_a() == Material.field_151586_h)
        {
            FluidUtil.tryPlaceFluid(null, worldIn, new FluidStack(FluidRegistry.WATER, Fluid.BUCKET_VOLUME), pos);
        }
        else if (block.func_176223_P().func_185904_a() == Material.field_151587_i)
        {
            FluidUtil.tryPlaceFluid(null, worldIn, new FluidStack(FluidRegistry.LAVA, Fluid.BUCKET_VOLUME), pos);
        }
        return false;
    }

    @SubscribeEvent(priority = EventPriority.LOW) // low priority so other mods can handle their stuff first
    public void onFillBucket(FillBucketEvent event)
    {
        if (event.getResult() != Event.Result.DEFAULT)
        {
            // event was already handled
            return;
        }

        // not for us to handle
        ItemStack emptyBucket = event.getEmptyBucket();
        if (emptyBucket == null ||
                !emptyBucket.func_77969_a(getEmpty()) ||
                (isNbtSensitive() && ItemStack.func_77970_a(emptyBucket, getEmpty())))
        {
            return;
        }

        // needs to target a block
        RayTraceResult target = event.getTarget();
        if (target == null || target.field_72313_a != RayTraceResult.Type.BLOCK)
        {
            return;
        }

        World world = event.getWorld();
        BlockPos pos = target.func_178782_a();

        ItemStack singleBucket = emptyBucket.func_77946_l();
        singleBucket.field_77994_a = 1;

        ItemStack filledBucket = FluidUtil.tryPickUpFluid(singleBucket, event.getEntityPlayer(), world, pos, target.field_178784_b);
        if (filledBucket != null)
        {
            event.setResult(Event.Result.ALLOW);
            event.setFilledBucket(filledBucket);
        }
        else
        {
            // cancel event, otherwise the vanilla minecraft ItemBucket would
            // convert it into a water/lava bucket depending on the blocks material
            event.setCanceled(true);
        }
    }

    public static ItemStack getFilledBucket(UniversalBucket item, Fluid fluid)
    {
        ItemStack stack = new ItemStack(item);
        item.fill(stack, new FluidStack(fluid, item.getCapacity()), true);
        return stack;
    }

  /* FluidContainer Management */

    @Override
    public FluidStack getFluid(ItemStack container)
    {
        return FluidStack.loadFluidStackFromNBT(container.func_77978_p());
    }

    @Override
    public int getCapacity(ItemStack container)
    {
        return getCapacity();
    }

    @Override
    public int fill(ItemStack container, FluidStack resource, boolean doFill)
    {
        // has to be exactly 1, must be handled from the caller
        if (container.field_77994_a != 1)
        {
            return 0;
        }

        // can only fill exact capacity
        if (resource == null || resource.amount < getCapacity())
        {
            return 0;
        }

        // already contains fluid?
        if (getFluid(container) != null)
        {
            return 0;
        }

        // registered in the registry?
        if (FluidRegistry.getBucketFluids().contains(resource.getFluid()))
        {
            // fill the container
            if (doFill)
            {
                NBTTagCompound tag = container.func_77978_p();
                if (tag == null)
                {
                    tag = new NBTTagCompound();
                }
                resource.writeToNBT(tag);
                container.func_77982_d(tag);
            }
            return getCapacity();
        }
        else if (resource.getFluid() == FluidRegistry.WATER)
        {
            if (doFill)
            {
                container.deserializeNBT(new ItemStack(Items.field_151131_as).serializeNBT());
            }
            return getCapacity();
        }
        else if (resource.getFluid() == FluidRegistry.LAVA)
        {
            if (doFill)
            {
                container.deserializeNBT(new ItemStack(Items.field_151129_at).serializeNBT());
            }
            return getCapacity();
        }

        return 0;
    }

    @Override
    public FluidStack drain(ItemStack container, int maxDrain, boolean doDrain)
    {
        // has to be exactly 1, must be handled from the caller
        if (container.field_77994_a != 1)
        {
            return null;
        }

        // can only drain everything at once
        if (maxDrain < getCapacity(container))
        {
            return null;
        }

        FluidStack fluidStack = getFluid(container);
        if (doDrain && fluidStack != null)
        {
            if(getEmpty() != null)
            {
                container.deserializeNBT(getEmpty().serializeNBT());
            }
            else {
                container.field_77994_a = 0;
            }
        }

        return fluidStack;
    }

    public int getCapacity()
    {
        return capacity;
    }

    public ItemStack getEmpty()
    {
        return empty;
    }

    public boolean isNbtSensitive()
    {
        return nbtSensitive;
    }

    @Override
    public ICapabilityProvider initCapabilities(ItemStack stack, NBTTagCompound nbt)
    {
        return new FluidBucketWrapper(stack);
    }
}
